@@ -42,6 +42,7 @@ var transformFuncs = map[string]transformFunc{
 	"day_of_week":          newTransformFuncDateTime(transformDayOfWeek),
 	"days_in_month":        newTransformFuncDateTime(transformDaysInMonth),
 	"deg":                  newTransformFuncOneArg(transformDeg),
+	"drop_common_labels":   transformDropCommonLabels,
 	"end":                  newTransformFuncZeroArgs(transformEnd),
 	"exp":                  newTransformFuncOneArg(transformExp),
 	"floor":                newTransformFuncOneArg(transformFloor),
@@ -658,6 +659,7 @@ func transformHistogramShare(tfa *transformFuncArg) ([]*timeseries, error) {
 		sort.Slice(xss, func(i, j int) bool {
 			return xss[i].le < xss[j].le
 		})
+		xss = mergeSameLE(xss)
 		dst := xss[0].ts
 		var tsLower, tsUpper *timeseries
 		if len(boundsLabel) > 0 {
@@ -821,8 +823,12 @@ func transformHistogramQuantiles(tfa *transformFuncArg) ([]*timeseries, error) {
 	tssOrig := args[len(args)-1]
 	// Calculate quantile individually per each phi.
 	var rvs []*timeseries
-	for _, phiArg := range phiArgs {
-		phiStr := fmt.Sprintf("%g", phiArg[0].Values[0])
+	for i, phiArg := range phiArgs {
+		phis, err := getScalar(phiArg, i)
+		if err != nil {
+			return nil, fmt.Errorf("cannot parse phi: %w", err)
+		}
+		phiStr := fmt.Sprintf("%g", phis[0])
 		tss := copyTimeseries(tssOrig)
 		tfaTmp := &transformFuncArg{
 			ec: tfa.ec,
@@ -934,6 +940,7 @@ func transformHistogramQuantile(tfa *transformFuncArg) ([]*timeseries, error) {
 		sort.Slice(xss, func(i, j int) bool {
 			return xss[i].le < xss[j].le
 		})
+		xss = mergeSameLE(xss)
 		dst := xss[0].ts
 		var tsLower, tsUpper *timeseries
 		if len(boundsLabel) > 0 {
@@ -993,18 +1000,57 @@ func groupLeTimeseries(tss []*timeseries) map[string][]leTimeseries {
 }
 
 func fixBrokenBuckets(i int, xss []leTimeseries) {
-	// Fix broken buckets.
-	// They are already sorted by le, so their values must be in ascending order,
+	// Buckets are already sorted by le, so their values must be in ascending order,
 	// since the next bucket includes all the previous buckets.
-	vPrev := float64(0)
-	for _, xs := range xss {
-		v := xs.ts.Values[i]
-		if v < vPrev || math.IsNaN(v) {
-			xs.ts.Values[i] = vPrev
-		} else {
-			vPrev = v
+	// If the next bucket has lower value than the current bucket,
+	// then the current bucket must be substituted with the next bucket value.
+	// See https://github.com/VictoriaMetrics/VictoriaMetrics/issues/2819
+	if len(xss) < 2 {
+		return
+	}
+	// Fill NaN in upper buckets with the first non-NaN value found in lower buckets.
+	for j := len(xss) - 1; j >= 0; j-- {
+		v := xss[j].ts.Values[i]
+		if !math.IsNaN(v) {
+			j++
+			for j < len(xss) {
+				xss[j].ts.Values[i] = v
+				j++
+			}
+			break
 		}
 	}
+	// Substitute lower bucket values with upper values if the lower values are NaN
+	// or are bigger than the upper bucket values.
+	vNext := xss[len(xss)-1].ts.Values[i]
+	for j := len(xss) - 2; j >= 0; j-- {
+		v := xss[j].ts.Values[i]
+		if math.IsNaN(v) || v > vNext {
+			xss[j].ts.Values[i] = vNext
+		} else {
+			vNext = v
+		}
+	}
+}
+
+func mergeSameLE(xss []leTimeseries) []leTimeseries {
+	// Merge buckets with identical le values.
+	// See https://github.com/VictoriaMetrics/VictoriaMetrics/pull/3225
+	xsDst := xss[0]
+	dst := xss[:1]
+	for j := 1; j < len(xss); j++ {
+		xs := xss[j]
+		if xs.le != xsDst.le {
+			dst = append(dst, xs)
+			xsDst = xs
+			continue
+		}
+		dstValues := xsDst.ts.Values
+		for k, v := range xs.ts.Values {
+			dstValues[k] += v
+		}
+	}
+	return dst
 }
 
 func transformHour(t time.Time) int {
@@ -1486,6 +1532,43 @@ func transformLabelMap(tfa *transformFuncArg) ([]*timeseries, error) {
 	return rvs, nil
 }
 
+func transformDropCommonLabels(tfa *transformFuncArg) ([]*timeseries, error) {
+	args := tfa.args
+	if len(args) < 1 {
+		return nil, fmt.Errorf(`not enough args; got %d; want at least %d`, len(args), 1)
+	}
+	rvs := args[0]
+	for _, tss := range args[1:] {
+		rvs = append(rvs, tss...)
+	}
+	m := make(map[string]map[string]int)
+	countLabel := func(name, value string) {
+		x := m[name]
+		if x == nil {
+			x = make(map[string]int)
+			m[name] = x
+		}
+		x[value]++
+	}
+	for _, ts := range rvs {
+		countLabel("__name__", string(ts.MetricName.MetricGroup))
+		for _, tag := range ts.MetricName.Tags {
+			countLabel(string(tag.Key), string(tag.Value))
+		}
+	}
+	for labelName, x := range m {
+		for _, count := range x {
+			if count != len(rvs) {
+				continue
+			}
+			for _, ts := range rvs {
+				ts.MetricName.RemoveTag(labelName)
+			}
+		}
+	}
+	return rvs, nil
+}
+
 func transformLabelCopy(tfa *transformFuncArg) ([]*timeseries, error) {
 	return transformLabelCopyExt(tfa, false)
 }
@@ -1677,8 +1760,10 @@ func transformLabelValue(tfa *transformFuncArg) ([]*timeseries, error) {
 			v = nan
 		}
 		values := ts.Values
-		for i := range values {
-			values[i] = v
+		for i, vOrig := range values {
+			if !math.IsNaN(vOrig) {
+				values[i] = v
+			}
 		}
 	}
 	// Do not remove timeseries with only NaN values, so `default` could be applied to them:
@@ -1788,7 +1873,9 @@ func transformLimitOffset(tfa *transformFuncArg) ([]*timeseries, error) {
 	if err != nil {
 		return nil, fmt.Errorf("cannot obtain offset arg: %w", err)
 	}
-	rvs := args[2]
+	// removeEmptySeries so offset will be calculated after empty series
+	// were filtered out.
+	rvs := removeEmptySeries(args[2])
 	if len(rvs) >= offset {
 		rvs = rvs[offset:]
 	}
@@ -2122,21 +2209,18 @@ func transformTimezoneOffset(tfa *transformFuncArg) ([]*timeseries, error) {
 	if err != nil {
 		return nil, fmt.Errorf("cannot get timezone name: %w", err)
 	}
-	tzOffset, err := getTimezoneOffset(tzString)
-	if err != nil {
-		return nil, fmt.Errorf("cannot get timezone offset for %q: %w", tzString, err)
-	}
-	rv := evalNumber(tfa.ec, float64(tzOffset))
-	return rv, nil
-}
-
-func getTimezoneOffset(tzString string) (int, error) {
 	loc, err := time.LoadLocation(tzString)
 	if err != nil {
-		return 0, fmt.Errorf("cannot load timezone %q: %w", tzString, err)
+		return nil, fmt.Errorf("cannot load timezone %q: %w", tzString, err)
 	}
-	_, tzOffset := time.Now().In(loc).Zone()
-	return tzOffset, nil
+
+	tss := evalNumber(tfa.ec, nan)
+	ts := tss[0]
+	for i, timestamp := range ts.Timestamps {
+		_, offset := time.Unix(timestamp/1000, 0).In(loc).Zone()
+		ts.Values[i] = float64(offset)
+	}
+	return tss, nil
 }
 
 func transformTime(tfa *transformFuncArg) ([]*timeseries, error) {
@@ -2289,9 +2373,9 @@ func removeCounterResetsMaybeNaNs(values []float64) {
 		d := v - prevValue
 		if d < 0 {
 			if (-d * 8) < prevValue {
-				// This is likely jitter from `Prometheus HA pairs`.
-				// Just substitute v with prevValue.
-				v = prevValue
+				// This is likely a partial counter reset.
+				// See https://github.com/VictoriaMetrics/VictoriaMetrics/issues/2787
+				correction += prevValue - v
 			} else {
 				correction += prevValue
 			}
