@@ -16,12 +16,13 @@ import (
 	"sync/atomic"
 	"time"
 
+	"github.com/VictoriaMetrics/VictoriaMetrics/lib/cgroup"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/logger"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/promauth"
 	"github.com/VictoriaMetrics/metrics"
 )
 
-var apiServerTimeout = flag.Duration("promscrape.kubernetes.apiServerTimeout", 30*time.Minute, "How frequently to reload the full state from Kuberntes API server")
+var apiServerTimeout = flag.Duration("promscrape.kubernetes.apiServerTimeout", 30*time.Minute, "How frequently to reload the full state from Kubernetes API server")
 
 // WatchEvent is a watch event returned from API server endpoints if `watch=1` query arg is set.
 //
@@ -45,7 +46,7 @@ type parseObjectFunc func(data []byte) (object, error)
 // parseObjectListFunc must parse objectList from the given r.
 type parseObjectListFunc func(r io.Reader) (map[string]object, ListMeta, error)
 
-// apiWatcher is used for watching for Kuberntes object changes and caching their latest states.
+// apiWatcher is used for watching for Kubernetes object changes and caching their latest states.
 type apiWatcher struct {
 	role string
 
@@ -54,7 +55,7 @@ type apiWatcher struct {
 
 	gw *groupWatcher
 
-	// swos contains per-urlWatcher maps of ScrapeWork objects for the given apiWatcher
+	// swosByURLWatcher contains per-urlWatcher maps of ScrapeWork objects for the given apiWatcher
 	swosByURLWatcher     map[*urlWatcher]map[string][]interface{}
 	swosByURLWatcherLock sync.Mutex
 
@@ -73,8 +74,9 @@ func newAPIWatcher(apiServer string, ac *promauth.Config, sdc *SDConfig, swcFunc
 		}
 	}
 	selectors := sdc.Selectors
-	proxyURL := sdc.ProxyURL.URL()
-	gw := getGroupWatcher(apiServer, ac, namespaces, selectors, proxyURL)
+	attachNodeMetadata := sdc.AttachMetadata.Node
+	proxyURL := sdc.ProxyURL.GetURL()
+	gw := getGroupWatcher(apiServer, ac, namespaces, selectors, attachNodeMetadata, proxyURL)
 	role := sdc.role()
 	return &apiWatcher{
 		role:             role,
@@ -89,25 +91,53 @@ func (aw *apiWatcher) mustStart() {
 	aw.gw.startWatchersForRole(aw.role, aw)
 }
 
+func (aw *apiWatcher) updateSwosCount(multiplier int, swosByKey map[string][]interface{}) {
+	n := 0
+	for _, swos := range swosByKey {
+		n += len(swos)
+	}
+	n *= multiplier
+	aw.swosCount.Add(n)
+}
+
 func (aw *apiWatcher) mustStop() {
 	aw.gw.unsubscribeAPIWatcher(aw)
 	aw.swosByURLWatcherLock.Lock()
 	for _, swosByKey := range aw.swosByURLWatcher {
-		aw.swosCount.Add(-len(swosByKey))
+		aw.updateSwosCount(-1, swosByKey)
 	}
 	aw.swosByURLWatcher = make(map[*urlWatcher]map[string][]interface{})
 	aw.swosByURLWatcherLock.Unlock()
 }
 
-func (aw *apiWatcher) reloadScrapeWorks(uw *urlWatcher, swosByKey map[string][]interface{}) {
+func (aw *apiWatcher) replaceScrapeWorks(uw *urlWatcher, swosByKey map[string][]interface{}) {
 	aw.swosByURLWatcherLock.Lock()
-	aw.swosCount.Add(len(swosByKey) - len(aw.swosByURLWatcher[uw]))
+	aw.updateSwosCount(-1, aw.swosByURLWatcher[uw])
+	aw.updateSwosCount(1, swosByKey)
 	aw.swosByURLWatcher[uw] = swosByKey
 	aw.swosByURLWatcherLock.Unlock()
 }
 
+func (aw *apiWatcher) updateScrapeWorks(uw *urlWatcher, swosByKey map[string][]interface{}) {
+	aw.swosByURLWatcherLock.Lock()
+	dst := aw.swosByURLWatcher[uw]
+	if dst == nil {
+		dst = make(map[string][]interface{})
+		aw.swosByURLWatcher[uw] = dst
+	}
+	for key, swos := range swosByKey {
+		aw.swosCount.Add(len(swos) - len(dst[key]))
+		if len(swos) == 0 {
+			delete(dst, key)
+		} else {
+			dst[key] = swos
+		}
+	}
+	aw.swosByURLWatcherLock.Unlock()
+}
+
 func (aw *apiWatcher) setScrapeWorks(uw *urlWatcher, key string, labels []map[string]string) {
-	swos := aw.getScrapeWorkObjectsForLabels(labels)
+	swos := getScrapeWorkObjectsForLabels(aw.swcFunc, labels)
 	aw.swosByURLWatcherLock.Lock()
 	swosByKey := aw.swosByURLWatcher[uw]
 	if swosByKey == nil {
@@ -115,10 +145,10 @@ func (aw *apiWatcher) setScrapeWorks(uw *urlWatcher, key string, labels []map[st
 		aw.swosByURLWatcher[uw] = swosByKey
 	}
 	aw.swosCount.Add(len(swos) - len(swosByKey[key]))
-	if len(swos) > 0 {
-		swosByKey[key] = swos
-	} else {
+	if len(swos) == 0 {
 		delete(swosByKey, key)
+	} else {
+		swosByKey[key] = swos
 	}
 	aw.swosByURLWatcherLock.Unlock()
 }
@@ -133,10 +163,11 @@ func (aw *apiWatcher) removeScrapeWorks(uw *urlWatcher, key string) {
 	aw.swosByURLWatcherLock.Unlock()
 }
 
-func (aw *apiWatcher) getScrapeWorkObjectsForLabels(labelss []map[string]string) []interface{} {
-	swos := make([]interface{}, 0, len(labelss))
+func getScrapeWorkObjectsForLabels(swcFunc ScrapeWorkConstructorFunc, labelss []map[string]string) []interface{} {
+	// Do not pre-allocate swos, since it is likely the swos will be empty because of relabeling
+	var swos []interface{}
 	for _, labels := range labelss {
-		swo := aw.swcFunc(labels)
+		swo := swcFunc(labels)
 		// The reflect check is needed because of https://mangatmodi.medium.com/go-check-nil-interface-the-right-way-d142776edef1
 		if swo != nil && !reflect.ValueOf(swo).IsNil() {
 			swos = append(swos, swo)
@@ -149,26 +180,19 @@ func (aw *apiWatcher) getScrapeWorkObjectsForLabels(labelss []map[string]string)
 func (aw *apiWatcher) getScrapeWorkObjects() []interface{} {
 	aw.gw.registerPendingAPIWatchers()
 
+	swos := make([]interface{}, 0, aw.swosCount.Get())
 	aw.swosByURLWatcherLock.Lock()
-	defer aw.swosByURLWatcherLock.Unlock()
-
-	size := 0
-	for _, swosByKey := range aw.swosByURLWatcher {
-		for _, swosLocal := range swosByKey {
-			size += len(swosLocal)
-		}
-	}
-	swos := make([]interface{}, 0, size)
 	for _, swosByKey := range aw.swosByURLWatcher {
 		for _, swosLocal := range swosByKey {
 			swos = append(swos, swosLocal...)
 		}
 	}
+	aw.swosByURLWatcherLock.Unlock()
 	return swos
 }
 
 // groupWatcher watches for Kubernetes objects on the given apiServer with the given namespaces,
-// selectors using the given client.
+// selectors and attachNodeMetadata using the given client.
 type groupWatcher struct {
 	// Old Kubernetes doesn't support /apis/networking.k8s.io/v1/, so /apis/networking.k8s.io/v1beta1/ must be used instead.
 	// This flag is used for automatic substitution of v1 API path with v1beta1 API path during requests to apiServer.
@@ -178,17 +202,19 @@ type groupWatcher struct {
 	// This flag is used for automatic substitution of v1 API path with v1beta1 API path during requests to apiServer.
 	useDiscoveryV1Beta1 uint32
 
-	apiServer     string
-	namespaces    []string
-	selectors     []Selector
-	getAuthHeader func() string
-	client        *http.Client
+	apiServer          string
+	namespaces         []string
+	selectors          []Selector
+	attachNodeMetadata bool
+
+	setHeaders func(req *http.Request)
+	client     *http.Client
 
 	mu sync.Mutex
 	m  map[string]*urlWatcher
 }
 
-func newGroupWatcher(apiServer string, ac *promauth.Config, namespaces []string, selectors []Selector, proxyURL *url.URL) *groupWatcher {
+func newGroupWatcher(apiServer string, ac *promauth.Config, namespaces []string, selectors []Selector, attachNodeMetadata bool, proxyURL *url.URL) *groupWatcher {
 	var proxy func(*http.Request) (*url.URL, error)
 	if proxyURL != nil {
 		proxy = http.ProxyURL(proxyURL)
@@ -204,26 +230,28 @@ func newGroupWatcher(apiServer string, ac *promauth.Config, namespaces []string,
 		Timeout: *apiServerTimeout,
 	}
 	return &groupWatcher{
-		apiServer:     apiServer,
-		getAuthHeader: ac.GetAuthHeader,
-		namespaces:    namespaces,
-		selectors:     selectors,
-		client:        client,
-		m:             make(map[string]*urlWatcher),
+		apiServer:          apiServer,
+		namespaces:         namespaces,
+		selectors:          selectors,
+		attachNodeMetadata: attachNodeMetadata,
+
+		setHeaders: func(req *http.Request) { ac.SetHeaders(req, true) },
+		client:     client,
+		m:          make(map[string]*urlWatcher),
 	}
 }
 
-func getGroupWatcher(apiServer string, ac *promauth.Config, namespaces []string, selectors []Selector, proxyURL *url.URL) *groupWatcher {
+func getGroupWatcher(apiServer string, ac *promauth.Config, namespaces []string, selectors []Selector, attachNodeMetadata bool, proxyURL *url.URL) *groupWatcher {
 	proxyURLStr := "<nil>"
 	if proxyURL != nil {
 		proxyURLStr = proxyURL.String()
 	}
-	key := fmt.Sprintf("apiServer=%s, namespaces=%s, selectors=%s, proxyURL=%s, authConfig=%s",
-		apiServer, namespaces, selectorsKey(selectors), proxyURLStr, ac.String())
+	key := fmt.Sprintf("apiServer=%s, namespaces=%s, selectors=%s, attachNodeMetadata=%v, proxyURL=%s, authConfig=%s",
+		apiServer, namespaces, selectorsKey(selectors), attachNodeMetadata, proxyURLStr, ac.String())
 	groupWatchersLock.Lock()
 	gw := groupWatchers[key]
 	if gw == nil {
-		gw = newGroupWatcher(apiServer, ac, namespaces, selectors, proxyURL)
+		gw = newGroupWatcher(apiServer, ac, namespaces, selectors, attachNodeMetadata, proxyURL)
 		groupWatchers[key] = gw
 	}
 	groupWatchersLock.Unlock()
@@ -250,18 +278,57 @@ var (
 	})
 )
 
-func (gw *groupWatcher) getObjectByRoleLocked(role, namespace, name string) object {
-	if gw == nil {
-		// this is needed for testing
+type swosByKeyWithLock struct {
+	mu        sync.Mutex
+	swosByKey map[string][]interface{}
+}
+
+func (gw *groupWatcher) getScrapeWorkObjectsByAPIWatcherLocked(objectsByKey map[string]object, awsMap map[*apiWatcher]struct{}) map[*apiWatcher]*swosByKeyWithLock {
+	if len(awsMap) == 0 {
 		return nil
 	}
+	swosByAPIWatcher := make(map[*apiWatcher]*swosByKeyWithLock, len(awsMap))
+	for aw := range awsMap {
+		swosByAPIWatcher[aw] = &swosByKeyWithLock{
+			swosByKey: make(map[string][]interface{}),
+		}
+	}
+
+	// Generate ScrapeWork objects in parallel on available CPU cores.
+	// This should reduce the time needed for their generation on systems with many CPU cores.
+	var wg sync.WaitGroup
+	limiterCh := make(chan struct{}, cgroup.AvailableCPUs())
+	for key, o := range objectsByKey {
+		labels := o.getTargetLabels(gw)
+		wg.Add(1)
+		limiterCh <- struct{}{}
+		go func(key string, labels []map[string]string) {
+			for aw, e := range swosByAPIWatcher {
+				swos := getScrapeWorkObjectsForLabels(aw.swcFunc, labels)
+				e.mu.Lock()
+				e.swosByKey[key] = swos
+				e.mu.Unlock()
+			}
+			wg.Done()
+			<-limiterCh
+		}(key, labels)
+	}
+	wg.Wait()
+	return swosByAPIWatcher
+}
+
+func (gw *groupWatcher) getObjectByRoleLocked(role, namespace, name string) object {
+	if role == "node" {
+		// Node objects have no namespace
+		namespace = ""
+	}
 	key := namespace + "/" + name
 	for _, uw := range gw.m {
 		if uw.role != role {
 			// Role mismatch
 			continue
 		}
-		if uw.namespace != "" && uw.namespace != namespace {
+		if namespace != "" && uw.namespace != "" && uw.namespace != namespace {
 			// Namespace mismatch
 			continue
 		}
@@ -278,6 +345,9 @@ func (gw *groupWatcher) startWatchersForRole(role string, aw *apiWatcher) {
 		gw.startWatchersForRole("pod", nil)
 		gw.startWatchersForRole("service", nil)
 	}
+	if gw.attachNodeMetadata && (role == "pod" || role == "endpoints" || role == "endpointslice") {
+		gw.startWatchersForRole("node", nil)
+	}
 	paths := getAPIPathsWithNamespaces(role, gw.namespaces, gw.selectors)
 	for _, path := range paths {
 		apiURL := gw.apiServer + path
@@ -295,14 +365,26 @@ func (gw *groupWatcher) startWatchersForRole(role string, aw *apiWatcher) {
 		if needStart {
 			uw.reloadObjects()
 			go uw.watchForUpdates()
-			if role == "endpoints" || role == "endpointslice" {
-				// Refresh endpoints and enpointslices targets in background, since they depend on other object types such as pod and service.
+			if role == "endpoints" || role == "endpointslice" || (gw.attachNodeMetadata && role == "pod") {
+				// Refresh targets in background, since they depend on other object types such as pod, service or node.
+				// This should guarantee that the ScrapeWork objects for these objects are properly updated
+				// as soon as the objects they depend on are updated.
 				// This should fix https://github.com/VictoriaMetrics/VictoriaMetrics/issues/1240 .
 				go func() {
+					const minSleepTime = 5 * time.Second
+					sleepTime := minSleepTime
 					for {
-						time.Sleep(5 * time.Second)
+						time.Sleep(sleepTime)
+						startTime := time.Now()
 						gw.mu.Lock()
-						uw.reloadScrapeWorksForAPIWatchersLocked(uw.aws)
+						if uw.needRecreateScrapeWorks {
+							uw.needRecreateScrapeWorks = false
+							uw.recreateScrapeWorksLocked(uw.objectsByKey, uw.aws)
+							sleepTime = time.Since(startTime)
+							if sleepTime < minSleepTime {
+								sleepTime = minSleepTime
+							}
+						}
 						gw.mu.Unlock()
 					}
 				}()
@@ -325,9 +407,7 @@ func (gw *groupWatcher) doRequest(requestURL string) (*http.Response, error) {
 	if err != nil {
 		logger.Fatalf("cannot create a request for %q: %s", requestURL, err)
 	}
-	if ah := gw.getAuthHeader(); ah != "" {
-		req.Header.Set("Authorization", ah)
-	}
+	gw.setHeaders(req)
 	resp, err := gw.client.Do(req)
 	if err != nil {
 		return nil, err
@@ -377,7 +457,7 @@ type urlWatcher struct {
 	// Batch registering saves CPU time needed for registering big number of Kubernetes objects
 	// shared among big number of scrape jobs, since per-object labels are generated only once
 	// for all the scrape jobs (each scrape job is associated with a single apiWatcher).
-	// See reloadScrapeWorksForAPIWatchersLocked for details.
+	// See registerPendingAPIWatchersLocked for details.
 	awsPending map[*apiWatcher]struct{}
 
 	// aws contains registered apiWatcher objects
@@ -386,6 +466,8 @@ type urlWatcher struct {
 	// objectsByKey contains the latest state for objects obtained from apiURL
 	objectsByKey map[string]object
 
+	needRecreateScrapeWorks bool
+
 	resourceVersion string
 
 	objectsCount          *metrics.Counter
@@ -433,15 +515,42 @@ func (uw *urlWatcher) registerPendingAPIWatchersLocked() {
 	if len(uw.awsPending) == 0 {
 		return
 	}
-	awsPending := make([]*apiWatcher, 0, len(uw.awsPending))
+	uw.recreateScrapeWorksLocked(uw.objectsByKey, uw.awsPending)
 	for aw := range uw.awsPending {
-		awsPending = append(awsPending, aw)
 		uw.aws[aw] = struct{}{}
 	}
-	uw.reloadScrapeWorksForAPIWatchersLocked(uw.awsPending)
+	awsPendingLen := len(uw.awsPending)
 	uw.awsPending = make(map[*apiWatcher]struct{})
-	metrics.GetOrCreateCounter(fmt.Sprintf(`vm_promscrape_discovery_kubernetes_subscribers{role=%q,status="working"}`, uw.role)).Add(len(awsPending))
-	metrics.GetOrCreateCounter(fmt.Sprintf(`vm_promscrape_discovery_kubernetes_subscribers{role=%q,status="pending"}`, uw.role)).Add(-len(awsPending))
+	metrics.GetOrCreateCounter(fmt.Sprintf(`vm_promscrape_discovery_kubernetes_subscribers{role=%q,status="working"}`, uw.role)).Add(awsPendingLen)
+	metrics.GetOrCreateCounter(fmt.Sprintf(`vm_promscrape_discovery_kubernetes_subscribers{role=%q,status="pending"}`, uw.role)).Add(-awsPendingLen)
+}
+
+func (uw *urlWatcher) recreateScrapeWorksLocked(objectsByKey map[string]object, awsMap map[*apiWatcher]struct{}) {
+	es := uw.gw.getScrapeWorkObjectsByAPIWatcherLocked(objectsByKey, awsMap)
+	for aw, e := range es {
+		swosByKey := e.swosByKey
+		for key, swos := range swosByKey {
+			if len(swos) == 0 {
+				delete(swosByKey, key)
+			}
+		}
+		aw.replaceScrapeWorks(uw, swosByKey)
+	}
+}
+
+func (uw *urlWatcher) updateScrapeWorksLocked(objectsByKey map[string]object, awsMap map[*apiWatcher]struct{}) {
+	es := uw.gw.getScrapeWorkObjectsByAPIWatcherLocked(objectsByKey, awsMap)
+	for aw, e := range es {
+		aw.updateScrapeWorks(uw, e.swosByKey)
+	}
+}
+
+func (uw *urlWatcher) removeScrapeWorksLocked(keys []string) {
+	for _, key := range keys {
+		for aw := range uw.aws {
+			aw.removeScrapeWorks(uw, key)
+		}
+	}
 }
 
 func (uw *urlWatcher) unsubscribeAPIWatcherLocked(aw *apiWatcher) {
@@ -462,6 +571,7 @@ func (uw *urlWatcher) reloadObjects() string {
 		return uw.resourceVersion
 	}
 
+	startTime := time.Now()
 	requestURL := uw.apiURL
 	resp, err := uw.gw.doRequest(requestURL)
 	if err != nil {
@@ -482,60 +592,49 @@ func (uw *urlWatcher) reloadObjects() string {
 	}
 
 	uw.gw.mu.Lock()
-	var updated, removed, added int
-	for key := range uw.objectsByKey {
-		if _, ok := objectsByKey[key]; ok {
-			updated++
+	objectsAdded := make(map[string]object)
+	objectsUpdated := make(map[string]object)
+	var objectsRemoved []string
+	for key, oPrev := range uw.objectsByKey {
+		o, ok := objectsByKey[key]
+		if ok {
+			if !reflect.DeepEqual(oPrev, o) {
+				objectsUpdated[key] = o
+			}
+			// Overwrite oPrev with o even if these objects are equal.
+			// This should free up memory associated with oPrev.
+			uw.objectsByKey[key] = o
 		} else {
-			removed++
+			objectsRemoved = append(objectsRemoved, key)
+			delete(uw.objectsByKey, key)
 		}
 	}
-	for key := range objectsByKey {
+	for key, o := range objectsByKey {
 		if _, ok := uw.objectsByKey[key]; !ok {
-			added++
+			objectsAdded[key] = o
+			uw.objectsByKey[key] = o
 		}
 	}
-	uw.objectsByKey = objectsByKey
-	uw.reloadScrapeWorksForAPIWatchersLocked(uw.aws)
+	uw.removeScrapeWorksLocked(objectsRemoved)
+	uw.updateScrapeWorksLocked(objectsUpdated, uw.aws)
+	uw.updateScrapeWorksLocked(objectsAdded, uw.aws)
+	uw.needRecreateScrapeWorks = false
+	if len(objectsRemoved) > 0 || len(objectsUpdated) > 0 || len(objectsAdded) > 0 {
+		uw.maybeUpdateDependedScrapeWorksLocked()
+	}
 	uw.gw.mu.Unlock()
 
-	uw.objectsUpdated.Add(updated)
-	uw.objectsRemoved.Add(removed)
-	uw.objectsAdded.Add(added)
-	uw.objectsCount.Add(added - removed)
+	uw.objectsUpdated.Add(len(objectsUpdated))
+	uw.objectsRemoved.Add(len(objectsRemoved))
+	uw.objectsAdded.Add(len(objectsAdded))
+	uw.objectsCount.Add(len(objectsAdded) - len(objectsRemoved))
 	uw.resourceVersion = metadata.ResourceVersion
 
-	logger.Infof("reloaded %d objects from %q; updated=%d, removed=%d, added=%d, resourceVersion=%q",
-		len(objectsByKey), requestURL, updated, removed, added, uw.resourceVersion)
+	logger.Infof("reloaded %d objects from %q in %.3fs; updated=%d, removed=%d, added=%d, resourceVersion=%q",
+		len(objectsByKey), requestURL, time.Since(startTime).Seconds(), len(objectsUpdated), len(objectsRemoved), len(objectsAdded), uw.resourceVersion)
 	return uw.resourceVersion
 }
 
-func (uw *urlWatcher) reloadScrapeWorksForAPIWatchersLocked(awsMap map[*apiWatcher]struct{}) {
-	if len(awsMap) == 0 {
-		return
-	}
-	aws := make([]*apiWatcher, 0, len(awsMap))
-	for aw := range awsMap {
-		aws = append(aws, aw)
-	}
-	swosByKey := make([]map[string][]interface{}, len(aws))
-	for i := range aws {
-		swosByKey[i] = make(map[string][]interface{})
-	}
-	for key, o := range uw.objectsByKey {
-		labels := o.getTargetLabels(uw.gw)
-		for i, aw := range aws {
-			swos := aw.getScrapeWorkObjectsForLabels(labels)
-			if len(swos) > 0 {
-				swosByKey[i][key] = swos
-			}
-		}
-	}
-	for i, aw := range aws {
-		aw.reloadScrapeWorks(uw, swosByKey[i])
-	}
-}
-
 // watchForUpdates watches for object updates starting from uw.resourceVersion and updates the corresponding objects to the latest state.
 //
 // See https://kubernetes.io/docs/reference/using-api/api-concepts/#efficient-detection-of-changes
@@ -597,7 +696,7 @@ func (uw *urlWatcher) watchForUpdates() {
 	}
 }
 
-// readObjectUpdateStream reads Kuberntes watch events from r and updates locally cached objects according to the received events.
+// readObjectUpdateStream reads Kubernetes watch events from r and updates locally cached objects according to the received events.
 func (uw *urlWatcher) readObjectUpdateStream(r io.Reader) error {
 	d := json.NewDecoder(r)
 	var we WatchEvent
@@ -613,21 +712,7 @@ func (uw *urlWatcher) readObjectUpdateStream(r io.Reader) error {
 			}
 			key := o.key()
 			uw.gw.mu.Lock()
-			if _, ok := uw.objectsByKey[key]; !ok {
-				// if we.Type == "MODIFIED" is expected condition after recovering from the bookmarked resourceVersion.
-				uw.objectsCount.Inc()
-				uw.objectsAdded.Inc()
-			} else {
-				// if we.Type == "ADDED" is expected condition after recovering from the bookmarked resourceVersion.
-				uw.objectsUpdated.Inc()
-			}
-			uw.objectsByKey[key] = o
-			if len(uw.aws) > 0 {
-				labels := o.getTargetLabels(uw.gw)
-				for aw := range uw.aws {
-					aw.setScrapeWorks(uw, key, labels)
-				}
-			}
+			uw.updateObjectLocked(key, o)
 			uw.gw.mu.Unlock()
 		case "DELETED":
 			o, err := uw.parseObject(we.Object)
@@ -636,14 +721,7 @@ func (uw *urlWatcher) readObjectUpdateStream(r io.Reader) error {
 			}
 			key := o.key()
 			uw.gw.mu.Lock()
-			if _, ok := uw.objectsByKey[key]; ok {
-				uw.objectsCount.Dec()
-				uw.objectsRemoved.Inc()
-				delete(uw.objectsByKey, key)
-			}
-			for aw := range uw.aws {
-				aw.removeScrapeWorks(uw, key)
-			}
+			uw.removeObjectLocked(key)
 			uw.gw.mu.Unlock()
 		case "BOOKMARK":
 			// See https://kubernetes.io/docs/reference/using-api/api-concepts/#watch-bookmarks
@@ -670,6 +748,69 @@ func (uw *urlWatcher) readObjectUpdateStream(r io.Reader) error {
 	}
 }
 
+func (uw *urlWatcher) updateObjectLocked(key string, o object) {
+	oPrev, ok := uw.objectsByKey[key]
+	// Overwrite oPrev with o even if these objects are equal.
+	// This should free up memory associated with oPrev.
+	uw.objectsByKey[key] = o
+	if !ok {
+		uw.objectsCount.Inc()
+		uw.objectsAdded.Inc()
+	} else {
+		if reflect.DeepEqual(oPrev, o) {
+			// Nothing to do, since the new object is equal to the previous one.
+			return
+		}
+		uw.objectsUpdated.Inc()
+	}
+	if len(uw.aws) > 0 {
+		labels := o.getTargetLabels(uw.gw)
+		for aw := range uw.aws {
+			aw.setScrapeWorks(uw, key, labels)
+		}
+	}
+	uw.maybeUpdateDependedScrapeWorksLocked()
+}
+
+func (uw *urlWatcher) removeObjectLocked(key string) {
+	if _, ok := uw.objectsByKey[key]; !ok {
+		return
+	}
+	uw.objectsCount.Dec()
+	uw.objectsRemoved.Inc()
+	delete(uw.objectsByKey, key)
+	for aw := range uw.aws {
+		aw.removeScrapeWorks(uw, key)
+	}
+	uw.maybeUpdateDependedScrapeWorksLocked()
+}
+
+func (uw *urlWatcher) maybeUpdateDependedScrapeWorksLocked() {
+	role := uw.role
+	attachNodeMetadata := uw.gw.attachNodeMetadata
+	if !(role == "pod" || role == "service" || (attachNodeMetadata && role == "node")) {
+		// Nothing to update
+		return
+	}
+	namespace := uw.namespace
+	for _, uwx := range uw.gw.m {
+		if namespace != "" && uwx.namespace != "" && uwx.namespace != namespace {
+			// Namespace mismatch
+			continue
+		}
+		if (role == "pod" || role == "service") && (uwx.role == "endpoints" || uwx.role == "endpointslice") {
+			// endpoints and endpointslice objects depend on pods and service objects
+			uwx.needRecreateScrapeWorks = true
+			continue
+		}
+		if attachNodeMetadata && role == "node" && (uwx.role == "pod" || uwx.role == "endpoints" || uwx.role == "endpointslice") {
+			// pod, endpoints and enpointslices objects depend on node objects if attachNodeMetadata is set
+			uwx.needRecreateScrapeWorks = true
+			continue
+		}
+	}
+}
+
 // Bookmark is a bookmark message from Kubernetes Watch API.
 // See https://kubernetes.io/docs/reference/using-api/api-concepts/#watch-bookmarks
 type Bookmark struct {
