@@ -44,7 +44,8 @@ func BenchmarkIndexDBAddTSIDs(b *testing.B) {
 	defer stopTestStorage(s)
 
 	dbName := nextIndexDBTableName()
-	db, err := openIndexDB(dbName, s, 0)
+	var isReadOnly uint32
+	db, err := openIndexDB(dbName, s, 0, &isReadOnly)
 	if err != nil {
 		b.Fatalf("cannot open indexDB: %s", err)
 	}
@@ -83,6 +84,7 @@ func BenchmarkIndexDBAddTSIDs(b *testing.B) {
 
 func benchmarkIndexDBAddTSIDs(db *indexDB, tsid *TSID, mn *MetricName, startOffset, recordsPerLoop int) {
 	var metricName []byte
+	var metricNameRaw []byte
 	is := db.getIndexSearch(noDeadline)
 	defer db.putIndexSearch(is)
 	for i := 0; i < recordsPerLoop; i++ {
@@ -92,7 +94,8 @@ func benchmarkIndexDBAddTSIDs(db *indexDB, tsid *TSID, mn *MetricName, startOffs
 		}
 		mn.sortTags()
 		metricName = mn.Marshal(metricName[:0])
-		if err := is.GetOrCreateTSIDByName(tsid, metricName); err != nil {
+		metricNameRaw = mn.marshalRaw(metricNameRaw[:0])
+		if err := is.GetOrCreateTSIDByName(tsid, metricName, metricNameRaw, 0); err != nil {
 			panic(fmt.Errorf("cannot insert record: %w", err))
 		}
 	}
@@ -105,7 +108,8 @@ func BenchmarkHeadPostingForMatchers(b *testing.B) {
 	defer stopTestStorage(s)
 
 	dbName := nextIndexDBTableName()
-	db, err := openIndexDB(dbName, s, 0)
+	var isReadOnly uint32
+	db, err := openIndexDB(dbName, s, 0, &isReadOnly)
 	if err != nil {
 		b.Fatalf("cannot open indexDB: %s", err)
 	}
@@ -119,7 +123,10 @@ func BenchmarkHeadPostingForMatchers(b *testing.B) {
 	// Fill the db with data as in https://github.com/prometheus/prometheus/blob/23c0299d85bfeb5d9b59e994861553a25ca578e5/tsdb/head_bench_test.go#L66
 	var mn MetricName
 	var metricName []byte
+	var metricNameRaw []byte
 	var tsid TSID
+	is := db.getIndexSearch(noDeadline)
+	defer db.putIndexSearch(is)
 	addSeries := func(kvs ...string) {
 		mn.Reset()
 		for i := 0; i < len(kvs); i += 2 {
@@ -127,20 +134,21 @@ func BenchmarkHeadPostingForMatchers(b *testing.B) {
 		}
 		mn.sortTags()
 		metricName = mn.Marshal(metricName[:0])
-		if err := db.createTSIDByName(&tsid, metricName); err != nil {
+		metricNameRaw = mn.marshalRaw(metricNameRaw[:0])
+		if err := is.createTSIDByName(&tsid, metricName, metricNameRaw, 0); err != nil {
 			b.Fatalf("cannot insert record: %s", err)
 		}
 	}
 	for n := 0; n < 10; n++ {
 		ns := strconv.Itoa(n)
 		for i := 0; i < 100000; i++ {
-			is := strconv.Itoa(i)
-			addSeries("i", is, "n", ns, "j", "foo")
+			ix := strconv.Itoa(i)
+			addSeries("i", ix, "n", ns, "j", "foo")
 			// Have some series that won't be matched, to properly test inverted matches.
-			addSeries("i", is, "n", ns, "j", "bar")
-			addSeries("i", is, "n", "0_"+ns, "j", "bar")
-			addSeries("i", is, "n", "1_"+ns, "j", "bar")
-			addSeries("i", is, "n", "2_"+ns, "j", "foo")
+			addSeries("i", ix, "n", ns, "j", "bar")
+			addSeries("i", ix, "n", "0_"+ns, "j", "bar")
+			addSeries("i", ix, "n", "1_"+ns, "j", "bar")
+			addSeries("i", ix, "n", "2_"+ns, "j", "foo")
 		}
 	}
 
@@ -157,7 +165,7 @@ func BenchmarkHeadPostingForMatchers(b *testing.B) {
 			MaxTimestamp: timestampFromTime(time.Now()),
 		}
 		for i := 0; i < b.N; i++ {
-			metricIDs, err := is.searchMetricIDs(tfss, tr, 2e9)
+			metricIDs, err := is.searchMetricIDs(nil, tfss, tr, 2e9)
 			if err != nil {
 				b.Fatalf("unexpected error in searchMetricIDs: %s", err)
 			}
@@ -280,7 +288,8 @@ func BenchmarkIndexDBGetTSIDs(b *testing.B) {
 	defer stopTestStorage(s)
 
 	dbName := nextIndexDBTableName()
-	db, err := openIndexDB(dbName, s, 0)
+	var isReadOnly uint32
+	db, err := openIndexDB(dbName, s, 0, &isReadOnly)
 	if err != nil {
 		b.Fatalf("cannot open indexDB: %s", err)
 	}
@@ -304,13 +313,15 @@ func BenchmarkIndexDBGetTSIDs(b *testing.B) {
 	}
 	var tsid TSID
 	var metricName []byte
+	var metricNameRaw []byte
 
 	is := db.getIndexSearch(noDeadline)
 	defer db.putIndexSearch(is)
 	for i := 0; i < recordsCount; i++ {
 		mn.sortTags()
 		metricName = mn.Marshal(metricName[:0])
-		if err := is.GetOrCreateTSIDByName(&tsid, metricName); err != nil {
+		metricNameRaw = mn.marshalRaw(metricName[:0])
+		if err := is.GetOrCreateTSIDByName(&tsid, metricName, metricNameRaw, 0); err != nil {
 			b.Fatalf("cannot insert record: %s", err)
 		}
 	}
@@ -321,6 +332,7 @@ func BenchmarkIndexDBGetTSIDs(b *testing.B) {
 	b.RunParallel(func(pb *testing.PB) {
 		var tsidLocal TSID
 		var metricNameLocal []byte
+		var metricNameLocalRaw []byte
 		mnLocal := mn
 		is := db.getIndexSearch(noDeadline)
 		defer db.putIndexSearch(is)
@@ -328,7 +340,8 @@ func BenchmarkIndexDBGetTSIDs(b *testing.B) {
 			for i := 0; i < recordsPerLoop; i++ {
 				mnLocal.sortTags()
 				metricNameLocal = mnLocal.Marshal(metricNameLocal[:0])
-				if err := is.GetOrCreateTSIDByName(&tsidLocal, metricNameLocal); err != nil {
+				metricNameLocalRaw = mnLocal.marshalRaw(metricNameLocalRaw[:0])
+				if err := is.GetOrCreateTSIDByName(&tsidLocal, metricNameLocal, metricNameLocalRaw, 0); err != nil {
 					panic(fmt.Errorf("cannot obtain tsid: %w", err))
 				}
 			}
