@@ -2,6 +2,7 @@ package storage
 
 import (
 	"bytes"
+	"errors"
 	"fmt"
 	"io"
 	"io/ioutil"
@@ -17,6 +18,7 @@ import (
 	"unsafe"
 
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/bloomfilter"
+	"github.com/VictoriaMetrics/VictoriaMetrics/lib/bytesutil"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/cgroup"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/decimal"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/encoding"
@@ -24,6 +26,8 @@ import (
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/fs"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/logger"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/memory"
+	"github.com/VictoriaMetrics/VictoriaMetrics/lib/querytracer"
+	"github.com/VictoriaMetrics/VictoriaMetrics/lib/snapshot"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/storagepacelimiter"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/timerpool"
 	"github.com/VictoriaMetrics/VictoriaMetrics/lib/uint64set"
@@ -166,10 +170,11 @@ func OpenStorage(path string, retentionMsecs int64, maxHourlySeries, maxDailySer
 	// See https://github.com/VictoriaMetrics/VictoriaMetrics/issues/1447 for details.
 	if fs.IsPathExist(s.cachePath + "/reset_cache_on_startup") {
 		logger.Infof("removing cache directory at %q, since it contains `reset_cache_on_startup` file...", s.cachePath)
-		var wg sync.WaitGroup
+		wg := getWaitGroup()
 		wg.Add(1)
 		fs.MustRemoveAllWithDoneCallback(s.cachePath, wg.Done)
 		wg.Wait()
+		putWaitGroup(wg)
 		logger.Infof("cache directory at %q has been successfully removed", s.cachePath)
 	}
 
@@ -256,7 +261,7 @@ func OpenStorage(path string, retentionMsecs int64, maxHourlySeries, maxDailySer
 
 	// Load data
 	tablePath := path + "/data"
-	tb, err := openTable(tablePath, s.getDeletedMetricIDs, retentionMsecs)
+	tb, err := openTable(tablePath, s.getDeletedMetricIDs, retentionMsecs, &s.isReadOnly)
 	if err != nil {
 		s.idb().MustClose()
 		return nil, fmt.Errorf("cannot open table at %q: %w", tablePath, err)
@@ -273,7 +278,7 @@ func OpenStorage(path string, retentionMsecs int64, maxHourlySeries, maxDailySer
 
 var maxTSIDCacheSize int
 
-// SetTSIDCacheSize overrides the default size of storage/tsid cahce
+// SetTSIDCacheSize overrides the default size of storage/tsid cache
 func SetTSIDCacheSize(size int) {
 	maxTSIDCacheSize = size
 }
@@ -316,7 +321,7 @@ func (s *Storage) CreateSnapshot() (string, error) {
 	s.snapshotLock.Lock()
 	defer s.snapshotLock.Unlock()
 
-	snapshotName := fmt.Sprintf("%s-%08X", time.Now().UTC().Format("20060102150405"), nextSnapshotIdx())
+	snapshotName := snapshot.NewName()
 	srcDir := s.path
 	dstDir := fmt.Sprintf("%s/snapshots/%s", srcDir, snapshotName)
 	if err := fs.MkdirAllFailIfExist(dstDir); err != nil {
@@ -362,7 +367,7 @@ func (s *Storage) CreateSnapshot() (string, error) {
 	srcMetadataDir := srcDir + "/metadata"
 	dstMetadataDir := dstDir + "/metadata"
 	if err := fs.CopyDirectory(srcMetadataDir, dstMetadataDir); err != nil {
-		return "", fmt.Errorf("cannot copy metadata: %s", err)
+		return "", fmt.Errorf("cannot copy metadata: %w", err)
 	}
 
 	fs.MustSyncPath(dstDir)
@@ -371,8 +376,6 @@ func (s *Storage) CreateSnapshot() (string, error) {
 	return snapshotName, nil
 }
 
-var snapshotNameRegexp = regexp.MustCompile("^[0-9]{14}-[0-9A-Fa-f]+$")
-
 // ListSnapshots returns sorted list of existing snapshots for s.
 func (s *Storage) ListSnapshots() ([]string, error) {
 	snapshotsPath := s.path + "/snapshots"
@@ -388,7 +391,7 @@ func (s *Storage) ListSnapshots() ([]string, error) {
 	}
 	snapshotNames := make([]string, 0, len(fnames))
 	for _, fname := range fnames {
-		if !snapshotNameRegexp.MatchString(fname) {
+		if err := snapshot.Validate(fname); err != nil {
 			continue
 		}
 		snapshotNames = append(snapshotNames, fname)
@@ -399,8 +402,8 @@ func (s *Storage) ListSnapshots() ([]string, error) {
 
 // DeleteSnapshot deletes the given snapshot.
 func (s *Storage) DeleteSnapshot(snapshotName string) error {
-	if !snapshotNameRegexp.MatchString(snapshotName) {
-		return fmt.Errorf("invalid snapshotName %q", snapshotName)
+	if err := snapshot.Validate(snapshotName); err != nil {
+		return fmt.Errorf("invalid snapshotName %q: %w", snapshotName, err)
 	}
 	snapshotPath := s.path + "/snapshots/" + snapshotName
 
@@ -417,10 +420,25 @@ func (s *Storage) DeleteSnapshot(snapshotName string) error {
 	return nil
 }
 
-var snapshotIdx = uint64(time.Now().UnixNano())
-
-func nextSnapshotIdx() uint64 {
-	return atomic.AddUint64(&snapshotIdx, 1)
+// DeleteStaleSnapshots deletes snapshot older than given maxAge
+func (s *Storage) DeleteStaleSnapshots(maxAge time.Duration) error {
+	list, err := s.ListSnapshots()
+	if err != nil {
+		return err
+	}
+	expireDeadline := time.Now().UTC().Add(-maxAge)
+	for _, snapshotName := range list {
+		t, err := snapshot.Time(snapshotName)
+		if err != nil {
+			return fmt.Errorf("cannot parse snapshot date from %q: %w", snapshotName, err)
+		}
+		if t.Before(expireDeadline) {
+			if err := s.DeleteSnapshot(snapshotName); err != nil {
+				return fmt.Errorf("cannot delete snapshot %q: %w", snapshotName, err)
+			}
+		}
+	}
+	return nil
 }
 
 func (s *Storage) idb() *indexDB {
@@ -493,6 +511,8 @@ type Metrics struct {
 	PrefetchedMetricIDsSize      uint64
 	PrefetchedMetricIDsSizeBytes uint64
 
+	NextRetentionSeconds uint64
+
 	IndexDBMetrics IndexDBMetrics
 	TableMetrics   TableMetrics
 }
@@ -583,6 +603,8 @@ func (s *Storage) UpdateMetrics(m *Metrics) {
 	m.PrefetchedMetricIDsSize += uint64(prefetchedMetricIDs.Len())
 	m.PrefetchedMetricIDsSizeBytes += uint64(prefetchedMetricIDs.SizeBytes())
 
+	m.NextRetentionSeconds = uint64(nextRetentionDuration(s.retentionMsecs).Seconds())
+
 	s.idb().UpdateMetrics(&m.IndexDBMetrics)
 	s.tb.UpdateMetrics(&m.TableMetrics)
 }
@@ -677,10 +699,12 @@ func (s *Storage) currHourMetricIDsUpdater() {
 	for {
 		select {
 		case <-s.stop:
-			s.updateCurrHourMetricIDs()
+			hour := fasttime.UnixHour()
+			s.updateCurrHourMetricIDs(hour)
 			return
 		case <-ticker.C:
-			s.updateCurrHourMetricIDs()
+			hour := fasttime.UnixHour()
+			s.updateCurrHourMetricIDs(hour)
 		}
 	}
 }
@@ -693,10 +717,12 @@ func (s *Storage) nextDayMetricIDsUpdater() {
 	for {
 		select {
 		case <-s.stop:
-			s.updateNextDayMetricIDs()
+			date := fasttime.UnixDate()
+			s.updateNextDayMetricIDs(date)
 			return
 		case <-ticker.C:
-			s.updateNextDayMetricIDs()
+			date := fasttime.UnixDate()
+			s.updateNextDayMetricIDs(date)
 		}
 	}
 }
@@ -706,7 +732,7 @@ func (s *Storage) mustRotateIndexDB() {
 	newTableName := nextIndexDBTableName()
 	idbNewPath := s.path + "/indexdb/" + newTableName
 	rotationTimestamp := fasttime.UnixTimestamp()
-	idbNew, err := openIndexDB(idbNewPath, s, rotationTimestamp)
+	idbNew, err := openIndexDB(idbNewPath, s, rotationTimestamp, &s.isReadOnly)
 	if err != nil {
 		logger.Panicf("FATAL: cannot create new indexDB at %q: %s", idbNewPath, err)
 	}
@@ -729,6 +755,27 @@ func (s *Storage) mustRotateIndexDB() {
 	// and slowly re-populate new idb with entries from the cache via maybeCreateIndexes().
 	// See https://github.com/VictoriaMetrics/VictoriaMetrics/issues/1401
 
+	// Flush metric id caches for the current and the previous hour,
+	// since they may contain entries missing in idbNew.
+	// This should prevent from missing data in queries when
+	// the following steps are performed for short -retentionPeriod (e.g. 1 day):
+	//
+	// 1. Add samples for some series between 3-4 UTC. These series are registered in currHourMetricIDs.
+	// 2. The indexdb rotation is performed at 4 UTC. currHourMetricIDs is moved to prevHourMetricIDs.
+	// 3. Continue adding samples for series from step 1 during time range 4-5 UTC.
+	//    These series are already registered in prevHourMetricIDs, so VM doesn't add per-day entries to the current indexdb.
+	// 4. Stop adding new samples for these series just before 5 UTC.
+	// 5. The next indexdb rotation is performed at 4 UTC next day.
+	//    The information about the series from step 5 disappears from indexdb, since the old indexdb from step 1 is deleted,
+	//    while the current indexdb doesn't contain information about the series.
+	//    So queries for the last 24 hours stop returning samples added at step 3.
+	// See https://github.com/VictoriaMetrics/VictoriaMetrics/issues/2698
+	s.pendingHourEntriesLock.Lock()
+	s.pendingHourEntries = &uint64set.Set{}
+	s.pendingHourEntriesLock.Unlock()
+	s.currHourMetricIDs.Store(&hourMetricIDs{})
+	s.prevHourMetricIDs.Store(&hourMetricIDs{})
+
 	// Flush dateMetricIDCache, so idbNew can be populated with fresh data.
 	s.dateMetricIDCache.Reset()
 
@@ -1020,6 +1067,14 @@ func (s *Storage) mustSaveCache(c *workingsetcache.Cache, info, name string) {
 // saveCacheLock prevents from data races when multiple concurrent goroutines save the same cache.
 var saveCacheLock sync.Mutex
 
+// SetRetentionTimezoneOffset sets the offset, which is used for calculating the time for indexdb rotation.
+// See https://github.com/VictoriaMetrics/VictoriaMetrics/pull/2574
+func SetRetentionTimezoneOffset(offset time.Duration) {
+	retentionTimezoneOffsetMsecs = offset.Milliseconds()
+}
+
+var retentionTimezoneOffsetMsecs int64
+
 func nextRetentionDuration(retentionMsecs int64) time.Duration {
 	// Round retentionMsecs to days. This guarantees that per-day inverted index works as expected.
 	retentionMsecs = ((retentionMsecs + msecPerDay - 1) / msecPerDay) * msecPerDay
@@ -1029,20 +1084,31 @@ func nextRetentionDuration(retentionMsecs int64) time.Duration {
 	// This should prevent from possible double deletion of indexdb
 	// due to time drift - see https://github.com/VictoriaMetrics/VictoriaMetrics/issues/248 .
 	deadline += 4 * 3600 * 1000
+	// The effect of time zone on retention period is moved out.
+	// See https://github.com/VictoriaMetrics/VictoriaMetrics/pull/2574
+	deadline -= retentionTimezoneOffsetMsecs
 	return time.Duration(deadline-t) * time.Millisecond
 }
 
-// SearchMetricNames returns metric names matching the given tfss on the given tr.
-func (s *Storage) SearchMetricNames(tfss []*TagFilters, tr TimeRange, maxMetrics int, deadline uint64) ([]MetricName, error) {
-	tsids, err := s.searchTSIDs(tfss, tr, maxMetrics, deadline)
+// SearchMetricNames returns marshaled metric names matching the given tfss on the given tr.
+//
+// The marshaled metric names must be unmarshaled via MetricName.UnmarshalString().
+func (s *Storage) SearchMetricNames(qt *querytracer.Tracer, tfss []*TagFilters, tr TimeRange, maxMetrics int, deadline uint64) ([]string, error) {
+	qt = qt.NewChild("search for matching metric names: filters=%s, timeRange=%s", tfss, &tr)
+	defer qt.Done()
+	tsids, err := s.searchTSIDs(qt, tfss, tr, maxMetrics, deadline)
 	if err != nil {
 		return nil, err
 	}
-	if err = s.prefetchMetricNames(tsids, deadline); err != nil {
+	if len(tsids) == 0 {
+		return nil, nil
+	}
+	if err = s.prefetchMetricNames(qt, tsids, deadline); err != nil {
 		return nil, err
 	}
 	idb := s.idb()
-	mns := make([]MetricName, 0, len(tsids))
+	metricNames := make([]string, 0, len(tsids))
+	metricNamesSeen := make(map[string]struct{}, len(tsids))
 	var metricName []byte
 	for i := range tsids {
 		if i&paceLimiterSlowIterationsMask == 0 {
@@ -1061,17 +1127,21 @@ func (s *Storage) SearchMetricNames(tfss []*TagFilters, tr TimeRange, maxMetrics
 			}
 			return nil, fmt.Errorf("error when searching metricName for metricID=%d: %w", metricID, err)
 		}
-		mns = mns[:len(mns)+1]
-		mn := &mns[len(mns)-1]
-		if err = mn.Unmarshal(metricName); err != nil {
-			return nil, fmt.Errorf("cannot unmarshal metricName=%q: %w", metricName, err)
+		if _, ok := metricNamesSeen[string(metricName)]; ok {
+			// The given metric name was already seen; skip it
+			continue
 		}
+		metricNames = append(metricNames, string(metricName))
+		metricNamesSeen[metricNames[len(metricNames)-1]] = struct{}{}
 	}
-	return mns, nil
+	qt.Printf("loaded %d metric names", len(metricNames))
+	return metricNames, nil
 }
 
 // searchTSIDs returns sorted TSIDs for the given tfss and the given tr.
-func (s *Storage) searchTSIDs(tfss []*TagFilters, tr TimeRange, maxMetrics int, deadline uint64) ([]TSID, error) {
+func (s *Storage) searchTSIDs(qt *querytracer.Tracer, tfss []*TagFilters, tr TimeRange, maxMetrics int, deadline uint64) ([]TSID, error) {
+	qt = qt.NewChild("search for matching tsids: filters=%s, timeRange=%s", tfss, &tr)
+	defer qt.Done()
 	// Do not cache tfss -> tsids here, since the caching is performed
 	// on idb level.
 
@@ -1092,6 +1162,7 @@ func (s *Storage) searchTSIDs(tfss []*TagFilters, tr TimeRange, maxMetrics int,
 		t := timerpool.Get(timeout)
 		select {
 		case searchTSIDsConcurrencyCh <- struct{}{}:
+			qt.Printf("wait in the queue because %d concurrent search requests are already performed", cap(searchTSIDsConcurrencyCh))
 			timerpool.Put(t)
 		case <-t.C:
 			timerpool.Put(t)
@@ -1100,7 +1171,7 @@ func (s *Storage) searchTSIDs(tfss []*TagFilters, tr TimeRange, maxMetrics int,
 				cap(searchTSIDsConcurrencyCh), timeout.Seconds())
 		}
 	}
-	tsids, err := s.idb().searchTSIDs(tfss, tr, maxMetrics, deadline)
+	tsids, err := s.idb().searchTSIDs(qt, tfss, tr, maxMetrics, deadline)
 	<-searchTSIDsConcurrencyCh
 	if err != nil {
 		return nil, fmt.Errorf("error when searching tsids: %w", err)
@@ -1118,8 +1189,11 @@ var (
 // prefetchMetricNames pre-fetches metric names for the given tsids into metricID->metricName cache.
 //
 // This should speed-up further searchMetricNameWithCache calls for metricIDs from tsids.
-func (s *Storage) prefetchMetricNames(tsids []TSID, deadline uint64) error {
+func (s *Storage) prefetchMetricNames(qt *querytracer.Tracer, tsids []TSID, deadline uint64) error {
+	qt = qt.NewChild("prefetch metric names for %d tsids", len(tsids))
+	defer qt.Done()
 	if len(tsids) == 0 {
+		qt.Printf("nothing to prefetch")
 		return nil
 	}
 	var metricIDs uint64Sorter
@@ -1132,8 +1206,10 @@ func (s *Storage) prefetchMetricNames(tsids []TSID, deadline uint64) error {
 		}
 		metricIDs = append(metricIDs, metricID)
 	}
+	qt.Printf("%d out of %d metric names must be pre-fetched", len(metricIDs), len(tsids))
 	if len(metricIDs) < 500 {
 		// It is cheaper to skip pre-fetching and obtain metricNames inline.
+		qt.Printf("skip pre-fetching metric names for low number of metrid ids=%d", len(metricIDs))
 		return nil
 	}
 	atomic.AddUint64(&s.slowMetricNameLoads, uint64(len(metricIDs)))
@@ -1180,6 +1256,7 @@ func (s *Storage) prefetchMetricNames(tsids []TSID, deadline uint64) error {
 	if err != nil {
 		return err
 	}
+	qt.Printf("pre-fetch metric names for %d metric ids", len(metricIDs))
 
 	// Store the pre-fetched metricIDs, so they aren't pre-fetched next time.
 	s.prefetchedMetricIDsLock.Lock()
@@ -1198,17 +1275,18 @@ func (s *Storage) prefetchMetricNames(tsids []TSID, deadline uint64) error {
 	}
 	s.prefetchedMetricIDs.Store(prefetchedMetricIDsNew)
 	s.prefetchedMetricIDsLock.Unlock()
+	qt.Printf("cache metric ids for pre-fetched metric names")
 	return nil
 }
 
 // ErrDeadlineExceeded is returned when the request times out.
 var ErrDeadlineExceeded = fmt.Errorf("deadline exceeded")
 
-// DeleteMetrics deletes all the metrics matching the given tfss.
+// DeleteSeries deletes all the series matching the given tfss.
 //
 // Returns the number of metrics deleted.
-func (s *Storage) DeleteMetrics(tfss []*TagFilters) (int, error) {
-	deletedCount, err := s.idb().DeleteTSIDs(tfss)
+func (s *Storage) DeleteSeries(qt *querytracer.Tracer, tfss []*TagFilters) (int, error) {
+	deletedCount, err := s.idb().DeleteTSIDs(qt, tfss)
 	if err != nil {
 		return deletedCount, fmt.Errorf("cannot delete tsids: %w", err)
 	}
@@ -1221,24 +1299,15 @@ func (s *Storage) DeleteMetrics(tfss []*TagFilters) (int, error) {
 	return deletedCount, nil
 }
 
-// SearchTagKeysOnTimeRange searches for tag keys on tr.
-func (s *Storage) SearchTagKeysOnTimeRange(tr TimeRange, maxTagKeys int, deadline uint64) ([]string, error) {
-	return s.idb().SearchTagKeysOnTimeRange(tr, maxTagKeys, deadline)
+// SearchLabelNamesWithFiltersOnTimeRange searches for label names matching the given tfss on tr.
+func (s *Storage) SearchLabelNamesWithFiltersOnTimeRange(qt *querytracer.Tracer, tfss []*TagFilters, tr TimeRange, maxLabelNames, maxMetrics int, deadline uint64) ([]string, error) {
+	return s.idb().SearchLabelNamesWithFiltersOnTimeRange(qt, tfss, tr, maxLabelNames, maxMetrics, deadline)
 }
 
-// SearchTagKeys searches for tag keys
-func (s *Storage) SearchTagKeys(maxTagKeys int, deadline uint64) ([]string, error) {
-	return s.idb().SearchTagKeys(maxTagKeys, deadline)
-}
-
-// SearchTagValuesOnTimeRange searches for tag values for the given tagKey on tr.
-func (s *Storage) SearchTagValuesOnTimeRange(tagKey []byte, tr TimeRange, maxTagValues int, deadline uint64) ([]string, error) {
-	return s.idb().SearchTagValuesOnTimeRange(tagKey, tr, maxTagValues, deadline)
-}
-
-// SearchTagValues searches for tag values for the given tagKey
-func (s *Storage) SearchTagValues(tagKey []byte, maxTagValues int, deadline uint64) ([]string, error) {
-	return s.idb().SearchTagValues(tagKey, maxTagValues, deadline)
+// SearchLabelValuesWithFiltersOnTimeRange searches for label values for the given labelName, filters and tr.
+func (s *Storage) SearchLabelValuesWithFiltersOnTimeRange(qt *querytracer.Tracer, labelName string, tfss []*TagFilters,
+	tr TimeRange, maxLabelValues, maxMetrics int, deadline uint64) ([]string, error) {
+	return s.idb().SearchLabelValuesWithFiltersOnTimeRange(qt, labelName, tfss, tr, maxLabelValues, maxMetrics, deadline)
 }
 
 // SearchTagValueSuffixes returns all the tag value suffixes for the given tagKey and tagValuePrefix on the given tr.
@@ -1246,14 +1315,15 @@ func (s *Storage) SearchTagValues(tagKey []byte, maxTagValues int, deadline uint
 // This allows implementing https://graphite-api.readthedocs.io/en/latest/api.html#metrics-find or similar APIs.
 //
 // If more than maxTagValueSuffixes suffixes is found, then only the first maxTagValueSuffixes suffixes is returned.
-func (s *Storage) SearchTagValueSuffixes(tr TimeRange, tagKey, tagValuePrefix []byte, delimiter byte, maxTagValueSuffixes int, deadline uint64) ([]string, error) {
-	return s.idb().SearchTagValueSuffixes(tr, tagKey, tagValuePrefix, delimiter, maxTagValueSuffixes, deadline)
+func (s *Storage) SearchTagValueSuffixes(qt *querytracer.Tracer, tr TimeRange, tagKey, tagValuePrefix string,
+	delimiter byte, maxTagValueSuffixes int, deadline uint64) ([]string, error) {
+	return s.idb().SearchTagValueSuffixes(qt, tr, tagKey, tagValuePrefix, delimiter, maxTagValueSuffixes, deadline)
 }
 
 // SearchGraphitePaths returns all the matching paths for the given graphite query on the given tr.
-func (s *Storage) SearchGraphitePaths(tr TimeRange, query []byte, maxPaths int, deadline uint64) ([]string, error) {
+func (s *Storage) SearchGraphitePaths(qt *querytracer.Tracer, tr TimeRange, query []byte, maxPaths int, deadline uint64) ([]string, error) {
 	query = replaceAlternateRegexpsWithGraphiteWildcards(query)
-	return s.searchGraphitePaths(tr, nil, query, maxPaths, deadline)
+	return s.searchGraphitePaths(qt, tr, nil, query, maxPaths, deadline)
 }
 
 // replaceAlternateRegexpsWithGraphiteWildcards replaces (foo|..|bar) with {foo,...,bar} in b and returns the new value.
@@ -1298,12 +1368,12 @@ func replaceAlternateRegexpsWithGraphiteWildcards(b []byte) []byte {
 	}
 }
 
-func (s *Storage) searchGraphitePaths(tr TimeRange, qHead, qTail []byte, maxPaths int, deadline uint64) ([]string, error) {
+func (s *Storage) searchGraphitePaths(qt *querytracer.Tracer, tr TimeRange, qHead, qTail []byte, maxPaths int, deadline uint64) ([]string, error) {
 	n := bytes.IndexAny(qTail, "*[{")
 	if n < 0 {
 		// Verify that qHead matches a metric name.
 		qHead = append(qHead, qTail...)
-		suffixes, err := s.SearchTagValueSuffixes(tr, nil, qHead, '.', 1, deadline)
+		suffixes, err := s.SearchTagValueSuffixes(qt, tr, "", bytesutil.ToUnsafeString(qHead), '.', 1, deadline)
 		if err != nil {
 			return nil, err
 		}
@@ -1318,7 +1388,7 @@ func (s *Storage) searchGraphitePaths(tr TimeRange, qHead, qTail []byte, maxPath
 		return []string{string(qHead)}, nil
 	}
 	qHead = append(qHead, qTail[:n]...)
-	suffixes, err := s.SearchTagValueSuffixes(tr, nil, qHead, '.', maxPaths, deadline)
+	suffixes, err := s.SearchTagValueSuffixes(qt, tr, "", bytesutil.ToUnsafeString(qHead), '.', maxPaths, deadline)
 	if err != nil {
 		return nil, err
 	}
@@ -1355,7 +1425,7 @@ func (s *Storage) searchGraphitePaths(tr TimeRange, qHead, qTail []byte, maxPath
 			continue
 		}
 		qHead = append(qHead[:qHeadLen], suffix...)
-		ps, err := s.searchGraphitePaths(tr, qHead, qTail, maxPaths, deadline)
+		ps, err := s.searchGraphitePaths(qt, tr, qHead, qTail, maxPaths, deadline)
 		if err != nil {
 			return nil, err
 		}
@@ -1426,39 +1496,6 @@ func getRegexpPartsForGraphiteQuery(q string) ([]string, string) {
 	}
 }
 
-// SearchTagEntries returns a list of (tagName -> tagValues)
-func (s *Storage) SearchTagEntries(maxTagKeys, maxTagValues int, deadline uint64) ([]TagEntry, error) {
-	idb := s.idb()
-	keys, err := idb.SearchTagKeys(maxTagKeys, deadline)
-	if err != nil {
-		return nil, fmt.Errorf("cannot search tag keys: %w", err)
-	}
-
-	// Sort keys for faster seeks below
-	sort.Strings(keys)
-
-	tes := make([]TagEntry, len(keys))
-	for i, key := range keys {
-		values, err := idb.SearchTagValues([]byte(key), maxTagValues, deadline)
-		if err != nil {
-			return nil, fmt.Errorf("cannot search values for tag %q: %w", key, err)
-		}
-		te := &tes[i]
-		te.Key = key
-		te.Values = values
-	}
-	return tes, nil
-}
-
-// TagEntry contains (tagName -> tagValues) mapping
-type TagEntry struct {
-	// Key is tagName
-	Key string
-
-	// Values contains all the values for Key.
-	Values []string
-}
-
 // GetSeriesCount returns the approximate number of unique time series.
 //
 // It includes the deleted series too and may count the same series
@@ -1467,9 +1504,9 @@ func (s *Storage) GetSeriesCount(deadline uint64) (uint64, error) {
 	return s.idb().GetSeriesCount(deadline)
 }
 
-// GetTSDBStatusWithFiltersForDate returns TSDB status data for /api/v1/status/tsdb with match[] filters.
-func (s *Storage) GetTSDBStatusWithFiltersForDate(tfss []*TagFilters, date uint64, topN int, deadline uint64) (*TSDBStatus, error) {
-	return s.idb().GetTSDBStatusWithFiltersForDate(tfss, date, topN, deadline)
+// GetTSDBStatus returns TSDB status data for /api/v1/status/tsdb
+func (s *Storage) GetTSDBStatus(qt *querytracer.Tracer, tfss []*TagFilters, date uint64, focusLabel string, topN, maxMetrics int, deadline uint64) (*TSDBStatus, error) {
+	return s.idb().GetTSDBStatus(qt, tfss, date, focusLabel, topN, maxMetrics, deadline)
 }
 
 // MetricRow is a metric to insert into storage.
@@ -1643,11 +1680,10 @@ var (
 //
 // The the MetricRow.Timestamp is used for registering the metric name starting from the given timestamp.
 // Th MetricRow.Value field is ignored.
-func (s *Storage) RegisterMetricNames(mrs []MetricRow) error {
-	var (
-		metricName []byte
-	)
-
+func (s *Storage) RegisterMetricNames(qt *querytracer.Tracer, mrs []MetricRow) error {
+	qt = qt.NewChild("registering %d series", len(mrs))
+	defer qt.Done()
+	var metricName []byte
 	var genTSID generationTSID
 	mn := GetMetricName()
 	defer PutMetricName(mn)
@@ -1658,64 +1694,37 @@ func (s *Storage) RegisterMetricNames(mrs []MetricRow) error {
 	for i := range mrs {
 		mr := &mrs[i]
 		if s.getTSIDFromCache(&genTSID, mr.MetricNameRaw) {
-			if genTSID.generation != idb.generation {
-				// The found entry is from the previous cache generation
-				// so attempt to re-populate the current generation with this entry.
-				// This is needed for https://github.com/VictoriaMetrics/VictoriaMetrics/issues/1401
-				created, err := idb.maybeCreateIndexes(&genTSID.TSID, mr.MetricNameRaw)
-				if err != nil {
-					return fmt.Errorf("cannot create indexes in the current indexdb: %w", err)
-				}
-				if created {
-					genTSID.generation = idb.generation
-					s.putTSIDToCache(&genTSID, mr.MetricNameRaw)
-				}
+			if err := s.registerSeriesCardinality(genTSID.TSID.MetricID, mr.MetricNameRaw); err != nil {
+				continue
+			}
+			if genTSID.generation == idb.generation {
+				// Fast path - mr.MetricNameRaw has been already registered in the current idb.
+				continue
 			}
-			// Fast path - mr.MetricNameRaw has been already registered.
-			continue
 		}
-
 		// Slow path - register mr.MetricNameRaw.
 		if err := mn.UnmarshalRaw(mr.MetricNameRaw); err != nil {
-			return fmt.Errorf("cannot register the metric because cannot unmarshal MetricNameRaw %q: %w", mr.MetricNameRaw, err)
+			return fmt.Errorf("cannot unmarshal MetricNameRaw %q: %w", mr.MetricNameRaw, err)
 		}
 		mn.sortTags()
 		metricName = mn.Marshal(metricName[:0])
-		if err := is.GetOrCreateTSIDByName(&genTSID.TSID, metricName); err != nil {
-			return fmt.Errorf("cannot register the metric because cannot create TSID for metricName %q: %w", metricName, err)
-		}
-		s.putTSIDToCache(&genTSID, mr.MetricNameRaw)
-
-		// Register the metric in per-day inverted index.
 		date := uint64(mr.Timestamp) / msecPerDay
-		metricID := genTSID.TSID.MetricID
-		if s.dateMetricIDCache.Has(date, metricID) {
-			// Fast path: the metric has been already registered in per-day inverted index
-			continue
-		}
-
-		// Slow path: acutally register the metric in per-day inverted index.
-		ok, err := is.hasDateMetricID(date, metricID)
-		if err != nil {
-			return fmt.Errorf("cannot register the metric in per-date inverted index because of error when locating (date=%d, metricID=%d) in database: %w",
-				date, metricID, err)
-		}
-		if !ok {
-			// The (date, metricID) entry is missing in the indexDB. Add it there.
-			if err := is.storeDateMetricID(date, metricID, mn); err != nil {
-				return fmt.Errorf("cannot register the metric in per-date inverted index because of error when storing (date=%d, metricID=%d) in database: %w",
-					date, metricID, err)
+		if err := is.GetOrCreateTSIDByName(&genTSID.TSID, metricName, mr.MetricNameRaw, date); err != nil {
+			if errors.Is(err, errSeriesCardinalityExceeded) {
+				continue
 			}
+			return fmt.Errorf("cannot create TSID for metricName %q: %w", metricName, err)
 		}
-		// The metric must be added to cache only after it has been successfully added to indexDB.
-		s.dateMetricIDCache.Set(date, metricID)
+		genTSID.generation = idb.generation
+		s.putTSIDToCache(&genTSID, mr.MetricNameRaw)
 	}
 	return nil
 }
 
 func (s *Storage) add(rows []rawRow, dstMrs []*MetricRow, mrs []MetricRow, precisionBits uint8) error {
 	idb := s.idb()
-	j := 0
+	is := idb.getIndexSearch(noDeadline)
+	defer idb.putIndexSearch(is)
 	var (
 		// These vars are used for speeding up bulk imports of multiple adjacent rows for the same metricName.
 		prevTSID          TSID
@@ -1728,6 +1737,7 @@ func (s *Storage) add(rows []rawRow, dstMrs []*MetricRow, mrs []MetricRow, preci
 
 	// Return only the first error, since it has no sense in returning all errors.
 	var firstWarn error
+	j := 0
 	for i := range mrs {
 		mr := &mrs[i]
 		if math.IsNaN(mr.Value) {
@@ -1771,26 +1781,26 @@ func (s *Storage) add(rows []rawRow, dstMrs []*MetricRow, mrs []MetricRow, preci
 			continue
 		}
 		if s.getTSIDFromCache(&genTSID, mr.MetricNameRaw) {
-			r.TSID = genTSID.TSID
-			if s.isSeriesCardinalityExceeded(r.TSID.MetricID, mr.MetricNameRaw) {
-				// Skip the row, since the limit on the number of unique series has been exceeded.
+			if err := s.registerSeriesCardinality(r.TSID.MetricID, mr.MetricNameRaw); err != nil {
 				j--
 				continue
 			}
+			r.TSID = genTSID.TSID
 			// Fast path - the TSID for the given MetricNameRaw has been found in cache and isn't deleted.
 			// There is no need in checking whether r.TSID.MetricID is deleted, since tsidCache doesn't
 			// contain MetricName->TSID entries for deleted time series.
-			// See Storage.DeleteMetrics code for details.
+			// See Storage.DeleteSeries code for details.
 			prevTSID = r.TSID
 			prevMetricNameRaw = mr.MetricNameRaw
 
 			if genTSID.generation != idb.generation {
-				// The found entry is from the previous cache generation
+				// The found entry is from the previous cache generation,
 				// so attempt to re-populate the current generation with this entry.
 				// This is needed for https://github.com/VictoriaMetrics/VictoriaMetrics/issues/1401
-				created, err := idb.maybeCreateIndexes(&genTSID.TSID, mr.MetricNameRaw)
+				date := uint64(r.Timestamp) / msecPerDay
+				created, err := is.maybeCreateIndexes(&genTSID.TSID, mr.MetricNameRaw, date)
 				if err != nil {
-					return fmt.Errorf("cannot create indexes in the current indexdb: %w", err)
+					return fmt.Errorf("cannot create indexes: %w", err)
 				}
 				if created {
 					genTSID.generation = idb.generation
@@ -1822,7 +1832,6 @@ func (s *Storage) add(rows []rawRow, dstMrs []*MetricRow, mrs []MetricRow, preci
 		sort.Slice(pendingMetricRows, func(i, j int) bool {
 			return string(pendingMetricRows[i].MetricName) < string(pendingMetricRows[j].MetricName)
 		})
-		is := idb.getIndexSearch(noDeadline)
 		prevMetricNameRaw = nil
 		var slowInsertsCount uint64
 		for i := range pendingMetricRows {
@@ -1838,78 +1847,79 @@ func (s *Storage) add(rows []rawRow, dstMrs []*MetricRow, mrs []MetricRow, preci
 				// Fast path - the current mr contains the same metric name as the previous mr, so it contains the same TSID.
 				// This path should trigger on bulk imports when many rows contain the same MetricNameRaw.
 				r.TSID = prevTSID
-				if s.isSeriesCardinalityExceeded(r.TSID.MetricID, mr.MetricNameRaw) {
-					// Skip the row, since the limit on the number of unique series has been exceeded.
-					j--
-					continue
-				}
 				continue
 			}
 			slowInsertsCount++
-			if err := is.GetOrCreateTSIDByName(&r.TSID, pmr.MetricName); err != nil {
+			date := uint64(r.Timestamp) / msecPerDay
+			if err := is.GetOrCreateTSIDByName(&r.TSID, pmr.MetricName, mr.MetricNameRaw, date); err != nil {
+				j--
+				if errors.Is(err, errSeriesCardinalityExceeded) {
+					continue
+				}
 				// Do not stop adding rows on error - just skip invalid row.
 				// This guarantees that invalid rows don't prevent
 				// from adding valid rows into the storage.
 				if firstWarn == nil {
 					firstWarn = fmt.Errorf("cannot obtain or create TSID for MetricName %q: %w", pmr.MetricName, err)
 				}
-				j--
 				continue
 			}
 			genTSID.generation = idb.generation
 			genTSID.TSID = r.TSID
 			s.putTSIDToCache(&genTSID, mr.MetricNameRaw)
+
 			prevTSID = r.TSID
 			prevMetricNameRaw = mr.MetricNameRaw
-			if s.isSeriesCardinalityExceeded(r.TSID.MetricID, mr.MetricNameRaw) {
-				// Skip the row, since the limit on the number of unique series has been exceeded.
-				j--
-				continue
-			}
 		}
-		idb.putIndexSearch(is)
 		putPendingMetricRows(pmrs)
 		atomic.AddUint64(&s.slowRowInserts, slowInsertsCount)
 	}
 	if firstWarn != nil {
-		logger.WithThrottler("storageAddRows", 5*time.Second).Warnf("warn occurred during rows addition: %s", firstWarn)
+		storageAddRowsLogger.Warnf("warn occurred during rows addition: %s", firstWarn)
 	}
 	dstMrs = dstMrs[:j]
 	rows = rows[:j]
 
-	var firstError error
-	if err := s.tb.AddRows(rows); err != nil {
-		firstError = fmt.Errorf("cannot add rows to table: %w", err)
+	err := s.updatePerDateData(rows, dstMrs)
+	if err != nil {
+		err = fmt.Errorf("cannot update per-date data: %w", err)
+	} else {
+		err = s.tb.AddRows(rows)
+		if err != nil {
+			err = fmt.Errorf("cannot add rows to table: %w", err)
+		}
 	}
-	if err := s.updatePerDateData(rows, dstMrs); err != nil && firstError == nil {
-		firstError = fmt.Errorf("cannot update per-date data: %w", err)
-	}
-	if firstError != nil {
-		return fmt.Errorf("error occurred during rows addition: %w", firstError)
+	if err != nil {
+		return fmt.Errorf("error occurred during rows addition: %w", err)
 	}
 	return nil
 }
 
-func (s *Storage) isSeriesCardinalityExceeded(metricID uint64, metricNameRaw []byte) bool {
+var storageAddRowsLogger = logger.WithThrottler("storageAddRows", 5*time.Second)
+
+func (s *Storage) registerSeriesCardinality(metricID uint64, metricNameRaw []byte) error {
 	if sl := s.hourlySeriesLimiter; sl != nil && !sl.Add(metricID) {
 		atomic.AddUint64(&s.hourlySeriesLimitRowsDropped, 1)
 		logSkippedSeries(metricNameRaw, "-storage.maxHourlySeries", sl.MaxItems())
-		return true
+		return errSeriesCardinalityExceeded
 	}
 	if sl := s.dailySeriesLimiter; sl != nil && !sl.Add(metricID) {
 		atomic.AddUint64(&s.dailySeriesLimitRowsDropped, 1)
 		logSkippedSeries(metricNameRaw, "-storage.maxDailySeries", sl.MaxItems())
-		return true
+		return errSeriesCardinalityExceeded
 	}
-	return false
+	return nil
 }
 
+var errSeriesCardinalityExceeded = fmt.Errorf("cannot create series because series cardinality limit exceeded")
+
 func logSkippedSeries(metricNameRaw []byte, flagName string, flagValue int) {
 	select {
 	case <-logSkippedSeriesTicker.C:
 		// Do not use logger.WithThrottler() here, since this will result in increased CPU load
 		// because of getUserReadableMetricName() calls per each logSkippedSeries call.
-		logger.Warnf("skip series %s because %s=%d reached", getUserReadableMetricName(metricNameRaw), flagName, flagValue)
+		userReadableMetricName := getUserReadableMetricName(metricNameRaw)
+		logger.Warnf("skip series %s because %s=%d reached", userReadableMetricName, flagName, flagValue)
 	default:
 	}
 }
@@ -2106,12 +2116,12 @@ func (s *Storage) updatePerDateData(rows []rawRow, mrs []*MetricRow) error {
 		ok, err := is.hasDateMetricID(date, metricID)
 		if err != nil {
 			if firstError == nil {
-				firstError = fmt.Errorf("error when locating (date=%d, metricID=%d) in database: %w", date, metricID, err)
+				firstError = fmt.Errorf("error when locating (date=%s, metricID=%d) in database: %w", dateToString(date), metricID, err)
 			}
 			continue
 		}
 		if !ok {
-			// The (date, metricID) entry is missing in the indexDB. Add it there.
+			// The (date, metricID) entry is missing in the indexDB. Add it there together with per-day indexes.
 			// It is OK if the (date, metricID) entry is added multiple times to db
 			// by concurrent goroutines.
 			if err := mn.UnmarshalRaw(dmid.mr.MetricNameRaw); err != nil {
@@ -2121,9 +2131,9 @@ func (s *Storage) updatePerDateData(rows []rawRow, mrs []*MetricRow) error {
 				continue
 			}
 			mn.sortTags()
-			if err := is.storeDateMetricID(date, metricID, mn); err != nil {
+			if err := is.createPerDayIndexes(date, metricID, mn); err != nil {
 				if firstError == nil {
-					firstError = fmt.Errorf("error when storing (date=%d, metricID=%d) in database: %w", date, metricID, err)
+					firstError = fmt.Errorf("error when storing per-date inverted index for (date=%s, metricID=%d): %w", dateToString(date), metricID, err)
 				}
 				continue
 			}
@@ -2345,8 +2355,7 @@ type byDateMetricIDEntry struct {
 	v    uint64set.Set
 }
 
-func (s *Storage) updateNextDayMetricIDs() {
-	date := fasttime.UnixDate()
+func (s *Storage) updateNextDayMetricIDs(date uint64) {
 	e := s.nextDayMetricIDs.Load().(*byDateMetricIDEntry)
 	s.pendingNextDayMetricIDsLock.Lock()
 	pendingMetricIDs := s.pendingNextDayMetricIDs
@@ -2360,6 +2369,11 @@ func (s *Storage) updateNextDayMetricIDs() {
 	// Slow path: union pendingMetricIDs with e.v
 	if e.date == date {
 		pendingMetricIDs.Union(&e.v)
+	} else {
+		// Do not add pendingMetricIDs from the previous day to the cyrrent day,
+		// since this may result in missing registration of the metricIDs in the per-day inverted index.
+		// See https://github.com/VictoriaMetrics/VictoriaMetrics/issues/3309
+		pendingMetricIDs = &uint64set.Set{}
 	}
 	eNew := &byDateMetricIDEntry{
 		date: date,
@@ -2368,13 +2382,13 @@ func (s *Storage) updateNextDayMetricIDs() {
 	s.nextDayMetricIDs.Store(eNew)
 }
 
-func (s *Storage) updateCurrHourMetricIDs() {
+func (s *Storage) updateCurrHourMetricIDs(hour uint64) {
 	hm := s.currHourMetricIDs.Load().(*hourMetricIDs)
 	s.pendingHourEntriesLock.Lock()
 	newMetricIDs := s.pendingHourEntries
 	s.pendingHourEntries = &uint64set.Set{}
 	s.pendingHourEntriesLock.Unlock()
-	hour := fasttime.UnixHour()
+
 	if newMetricIDs.Len() == 0 && hm.hour == hour {
 		// Fast path: nothing to update.
 		return
@@ -2389,7 +2403,12 @@ func (s *Storage) updateCurrHourMetricIDs() {
 		m = &uint64set.Set{}
 		isFull = true
 	}
-	m.Union(newMetricIDs)
+	if hour%24 != 0 {
+		// Do not add pending metricIDs from the previous hour to the current hour on the next day,
+		// since this may result in missing registration of the metricIDs in the per-day inverted index.
+		// See https://github.com/VictoriaMetrics/VictoriaMetrics/issues/3309
+		m.Union(newMetricIDs)
+	}
 	hmNew := &hourMetricIDs{
 		m:      m,
 		hour:   hour,
@@ -2484,12 +2503,12 @@ func (s *Storage) openIndexDBTables(path string) (curr, prev *indexDB, err error
 	// Open the last two tables.
 	currPath := path + "/" + tableNames[len(tableNames)-1]
 
-	curr, err = openIndexDB(currPath, s, 0)
+	curr, err = openIndexDB(currPath, s, 0, &s.isReadOnly)
 	if err != nil {
 		return nil, nil, fmt.Errorf("cannot open curr indexdb table at %q: %w", currPath, err)
 	}
 	prevPath := path + "/" + tableNames[len(tableNames)-2]
-	prev, err = openIndexDB(prevPath, s, 0)
+	prev, err = openIndexDB(prevPath, s, 0, &s.isReadOnly)
 	if err != nil {
 		curr.MustClose()
 		return nil, nil, fmt.Errorf("cannot open prev indexdb table at %q: %w", prevPath, err)
