@@ -17,7 +17,7 @@ const (
 	whole     = 2
 )
 
-const defaultExpireDuration = 20 * time.Minute
+const defaultExpireDuration = 10 * time.Minute
 
 // Cache is a cache for working set entries.
 //
@@ -144,25 +144,18 @@ func (c *Cache) expirationWatcher(expireDuration time.Duration) {
 			c.mu.Unlock()
 			return
 		}
-		// Expire prev cache and create fresh curr cache with the same capacity.
-		// Do not reuse prev cache, since it can occupy too big amounts of memory.
+		// Reset prev cache and swap it with the curr cache.
 		prev := c.prev.Load().(*fastcache.Cache)
-		prev.Reset()
 		curr := c.curr.Load().(*fastcache.Cache)
-		var cs fastcache.Stats
-		curr.UpdateStats(&cs)
 		c.prev.Store(curr)
-		// Use c.maxBytes/2 instead of cs.MaxBytesSize for creating new cache,
-		// since cs.MaxBytesSize may not match c.maxBytes/2, so the created cache
-		// couldn't be loaded from file with c.maxBytes/2 limit after saving with cs.MaxBytesSize size.
-		curr = fastcache.New(c.maxBytes / 2)
-		c.curr.Store(curr)
+		prev.Reset()
+		c.curr.Store(prev)
 		c.mu.Unlock()
 	}
 }
 
 func (c *Cache) cacheSizeWatcher() {
-	t := time.NewTicker(time.Minute)
+	t := time.NewTicker(1500 * time.Millisecond)
 	defer t.Stop()
 
 	var maxBytesSize uint64
@@ -199,9 +192,9 @@ func (c *Cache) cacheSizeWatcher() {
 	c.mu.Lock()
 	c.setMode(switching)
 	prev := c.prev.Load().(*fastcache.Cache)
-	prev.Reset()
 	curr := c.curr.Load().(*fastcache.Cache)
 	c.prev.Store(curr)
+	prev.Reset()
 	// use c.maxBytes instead of maxBytesSize*2 for creating new cache, since otherwise the created cache
 	// couldn't be loaded from file with c.maxBytes limit after saving with maxBytesSize*2 limit.
 	c.curr.Store(fastcache.New(c.maxBytes))
@@ -224,8 +217,8 @@ func (c *Cache) cacheSizeWatcher() {
 	c.mu.Lock()
 	c.setMode(whole)
 	prev = c.prev.Load().(*fastcache.Cache)
-	prev.Reset()
 	c.prev.Store(fastcache.New(1024))
+	prev.Reset()
 	c.mu.Unlock()
 }
 
